import os
from PIL import Image, ImageDraw
from robot.libraries.BuiltIn import BuiltIn
from .config import *


class Browser:
    def _get_current_browser(self):
        if hasattr(BuiltIn().get_library_instance(
                'SeleniumLibrary'), 'driver'):
            return BuiltIn().get_library_instance('SeleniumLibrary').driver
        else:
            return BuiltIn().get_library_instance('SeleniumLibrary')._current_browser()

    def _get_element_coordinates(self, element_locator):
        browser = self._get_current_browser()
        element = browser.find_element_by_xpath(element_locator)
        location = element.location
        size = element.size
        return location, size

    def _exclude_elements(self, locator, screenshot, excluded_locators):
        if isinstance(excluded_locators, str):
            excluded_locators = [excluded_locators, ]

        with Image.open(screenshot) as image:
            element_location, element_size = self._get_element_coordinates(
                locator)
            for item in excluded_locators:
                excluded_element_location, excluded_element_size = self._get_element_coordinates(
                    item)
                x1 = excluded_element_location['x'] - element_location['x']
                y1 = excluded_element_location['y'] - element_location['y']
                x2 = x1 + excluded_element_size['width']
                y2 = y1 + excluded_element_size['height']
                draw = ImageDraw.Draw(image)
                draw.rectangle([(x1, y1), (x2, y2)], fill="lightgrey")
            image.save(screenshot)

    def _save_screenshot(self, temp_dir, locator, name,
                         excluded_locators, trim):
        browser = self._get_current_browser()
        element = browser.find_element_by_xpath(locator)

        screenshot_path = os.path.join(temp_dir, name)
        saved_screenshot = element.screenshot(screenshot_path)
        if not saved_screenshot:
            raise Exception('Failed to save screenshot')

        if excluded_locators:
            self._exclude_elements(locator, screenshot_path, excluded_locators)

        if trim:
            im = Image.open(screenshot_path)
            width, height = im.size
            cropped_image = im.crop(
                (0 + int(trim),
                 0 + int(trim),
                    width - int(trim),
                    height - int(trim)))
            cropped_image.save(screenshot_path)

        return screenshot_path



import os
from mega import Mega
from robot.libraries.BuiltIn import _Variables as Variables
from .config import *


class MegaBaseLine:
    """
    Works with baseline placed on mega.nz
    """

    def __init__(self):
        mega_account_email = Variables().get_variable_value("${MEGA_EMAIL}")
        mega_account_passw = Variables().get_variable_value("${MEGA_PASSW}")

        self.mega = Mega()
        self.m = self.mega.login(mega_account_email, mega_account_passw)

    def get(self, temp_local_baseline_path, remote_mega_dir, image_expected):
        """
        Gets previously saved screenshot from Mega
        :param temp_local_baseline_path: Absolute path to local temp directory to download file to.
        :param remote_mega_dir: Name of the directory with a baseline on Mega
        :param image_expected: Name of the screenshot file with expected result.
        :return: Absolute path to downloaded baseline file
        """
        mega_dir = self.m.find(remote_mega_dir, exclude_deleted=True)
        files_in_target_folder = self.m.get_files_in_node(mega_dir[0])
        found_baseline_screenshot = False

        for file_id in files_in_target_folder.keys():
            if files_in_target_folder[file_id]['a']['n'] == image_expected:
                found_baseline_screenshot = True
                file_obj = (file_id, files_in_target_folder[file_id])
                self.m.download(
                    file_obj,
                    dest_path=temp_local_baseline_path,
                    dest_filename=f"expected-{image_expected}")
                break
        if found_baseline_screenshot:
            return os.path.join(temp_local_baseline_path,
                                f"expected-{image_expected}")
        else:
            print(
                f"Not found baseline screenshot '{image_expected}' in the dir '{remote_mega_dir}'")
            return None

    def save(self, temp_local_baseline_screenshot_path, remote_mega_dir):
        """
        Save new baseline to Mega
        :param temp_local_baseline_screenshot_path: Absolute path to local file to be saved in Mega.
        :param remote_mega_dir: Name of the directory with a baseline on Mega
        """
        mega_dir = self.m.find(remote_mega_dir, exclude_deleted=True)
        if mega_dir:
            files_in_target_folder = self.m.get_files_in_node(mega_dir[0])
            if len(files_in_target_folder) == 0:
                self.m.upload(
                    temp_local_baseline_screenshot_path,
                    dest=mega_dir[0])
            else:
                existing_file_nodes = files_in_target_folder.keys()
                for file_uuid in existing_file_nodes:
                    if files_in_target_folder[file_uuid]['a']['n'] == os.path.split(
                            temp_local_baseline_screenshot_path)[1]:
                        self.m.delete(file_uuid)
                        break
                self.m.upload(
                    temp_local_baseline_screenshot_path,
                    dest=mega_dir[0])
        else:
            raise Exception(f"Not found '{remote_mega_dir}' in MEGA")


class LocalBaseline:
    """
    Works with baseline placed on local machine
    """
    pass


class FtpServerBaseline:
    """
    Works with baseline placed on a remote server via FTP
    """
    pass



import os
from PIL import Image, ImageDraw
from robot.libraries.BuiltIn import BuiltIn
from .artifacts import Artifacts
from .config import *


class Comparator:
    def _region_analyze(self, image, x, y, width, height, factor):
        region_status = 0
        for x_cord in range(x, x + width):
            for y_cord in range(y, y + height):
                try:
                    pixel = image.getpixel((x_cord, y_cord))
                    region_status += int(sum(pixel) / 3)
                except BaseException:
                    return None
        return region_status // factor

    def _analyze(self, image_actual, image_expected, cols, rows, factor):
        """
        Analyzes two images and returns joined screenshot (actual result/expected result/diff) if images differs.
        :param image_actual: Path to screenshot with actual result.
        :param image_expected: File name of expected result. This file is searched in BASELINE path
        :param cols: qty of columns to compare images
        :param rows: qty of rows to compare images
        :return: None if images are identical, or path to saved file (actual/expected/diff) otherwise.
        """
        actual = Image.open(image_actual)
        expected = Image.open(image_expected)

        width, height = actual.size
        block_width = width // int(cols)
        block_height = height // int(rows)

        has_diff = False

        for x in range(0, width, block_width + 1):
            for y in range(0, height, block_height + 1):
                region_actual = self._region_analyze(
                    actual, x, y, block_width, block_height, factor)
                region_expected = self._region_analyze(
                    expected, x, y, block_width, block_height, factor)

                if region_actual != region_expected:
                    has_diff = True
                    draw = ImageDraw.Draw(actual)
                    draw.rectangle(
                        (x - 1, y - 1, x + block_width, y + block_height), outline="red")

        if has_diff:
            image_diff = os.path.join(
                os.path.split(image_actual)[0], f"diff-{os.path.split(image_actual)[1]}")

            actual.save(image_diff)
            joined_actual_expected_diff_path = Artifacts()._join_screenshots(
                image_actual, image_expected, image_diff)

            return joined_actual_expected_diff_path
        else:
            BuiltIn().log_to_console(
                "Images are identical on: FACTOR {f}, COLS {c}, ROWS {r}".format(f=factor, c=cols, r=rows))

            return None



import os
import string
import random
import shutil
from PIL import Image, ImageDraw, ImageFont
from robot.libraries.BuiltIn import _Variables as Variables
from .config import *


class Artifacts:
    def _create_temp_dir(self):
        temp_dir_path = os.path.join(
            Variables().get_variable_value("${OUTPUT DIR}"),
            f'temp-{"".join(random.choices(string.ascii_lowercase + string.digits, k=10))}')
        if not os.path.exists(temp_dir_path):
            os.makedirs(temp_dir_path)
            return temp_dir_path

    def _delete_temp_dir(self, temp_dir_path):
        if os.path.exists(temp_dir_path):
            shutil.rmtree(temp_dir_path)

    def _add_diff_file_to_report(self, source):
        source_file_name = os.path.split(source)[1]
        destination = os.path.join(
            Variables().get_variable_value("${OUTPUT DIR}"),
            DIR_FOR_SCREENSHOTS_IN_REPORT)
        if not os.path.exists(destination):
            os.makedirs(destination)

        destination_file = f'{"".join(random.choices(string.ascii_lowercase + string.digits, k=10))}-{source_file_name}'
        destination_file_path = os.path.join(destination, destination_file)
        shutil.copy(source, destination_file_path)
        return destination_file_path

    def _join_screenshots(self, image_actual, image_expected, image_diff):
        images = [
            Image.open(x) for x in [
                image_actual,
                image_expected,
                image_diff]]
        widths, heights = zip(*(i.size for i in images))

        total_width = sum(widths) + 20
        max_height = max(heights) + 30

        report_im = Image.new(
            'RGB', (total_width, max_height), color="dimgrey")
        report_im.paste(images[0], (5, 25))
        report_im.paste(images[1], (images[0].size[0] + 10, 25))
        report_im.paste(
            images[2],
            (images[0].size[0] +
             images[1].size[0] +
                15,
                25))
        title_font = "/usr/share/fonts/truetype/liberation/LiberationMono-Bold.ttf"
        if os.path.exists(title_font):
            titles = ImageDraw.Draw(report_im)
            titles.text((5, 5), "ACTUAL", fill="yellow",
                        font=ImageFont.truetype(title_font, 16))
            titles.text(
                (images[0].size[0] + 10,
                 5),
                "EXPECTED",
                fill="lawngreen",
                font=ImageFont.truetype(
                    title_font,
                    16))
            titles.text(
                (images[0].size[0] +
                 images[1].size[0] +
                    15,
                    5),
                "DIFF",
                fill="red",
                font=ImageFont.truetype(
                    title_font,
                    16))
        else:
            titles = ImageDraw.Draw(report_im)
            titles.text((5, 5), "ACTUAL", fill="yellow")
            titles.text((images[0].size[0] + 10, 5),
                        "EXPECTED", fill="lawngreen")
            titles.text(
                (images[0].size[0] +
                 images[1].size[0] +
                    15,
                    5),
                "DIFF",
                fill="red")

        report_im.save(image_diff)
        return image_diff



import os
from robot.libraries.BuiltIn import _Variables as Variables
from robot.libraries.BuiltIn import BuiltIn
from robot.api.deco import keyword, library
from .baseline import MegaBaseLine, LocalBaseline, FtpServerBaseline
from .comparator import Comparator
from .browser import Browser
from .artifacts import Artifacts
from .config import *


@library(scope='GLOBAL', auto_keywords=False)
class VisualAssertLibrary:

    @keyword
    def assert_screenshot(self, message, element_locator, image_expected,
                          exclude=None, cols=20, rows=20, factor=1000, trim=None):
        """
        Keyword to assert web element screenshot.
        :param message: Text to be displayed in report if test failed.
        :param element_locator: xpath to find the element that under test
        :param image_expected: path to the base line screenshot with expected result
        :param exclude: xpath to element to be excluded from comparing, xpath string or list of strings
        :param cols: number of columns to split images for comparison
        :param rows: number of rows to split images for comparison
        :param factor: accuracy of comparing screenshots
        :param trim: trim passed number of pixels from top/bottom/left/right
        :return: absolute path to joined diff image (actual result / expected result / diff) if test failed (screenshots
                 differ) or None if test passed (screenshots are identical)
        """
        if ".png" not in image_expected:
            image_expected = image_expected + ".png"

        # Use baseline if exist or save new baseline otherwise:
        if Variables().get_variable_value("${MEGA_BASELINE_DIR_NAME}"):
            temp_dir = Artifacts()._create_temp_dir()
            BuiltIn().log_to_console(f"\n1. Create temp dir {temp_dir}")

            BuiltIn().log_to_console("2. Search baseline with expected result")
            remote_mega_dir_name = Variables().get_variable_value(
                "${MEGA_BASELINE_DIR_NAME}")

            screenshot_actual = Browser()._save_screenshot(
                temp_dir, element_locator, image_expected, exclude, trim)

            screenshot_expected = MegaBaseLine().get(
                temp_dir, remote_mega_dir_name, image_expected)
            if screenshot_expected:
                diff_screenshot = Comparator()._analyze(
                    screenshot_actual, screenshot_expected, cols, rows, factor)
                if diff_screenshot:
                    BuiltIn().log_to_console(
                        "3. Test failed. Add diff screenshot to report")
                    report_diff_screenshot = Artifacts()._add_diff_file_to_report(diff_screenshot)
                    BuiltIn().log_to_console(f"4. Delete temp dir {temp_dir}")
                    Artifacts()._delete_temp_dir(temp_dir)
                    self._fail_visual_test(
                        message, image_expected, report_diff_screenshot)
                else:
                    BuiltIn().log_to_console("3. Test Pass")
                    BuiltIn().log_to_console(f"4. Delete temp dir {temp_dir}")
                    Artifacts()._delete_temp_dir(temp_dir)

            else:
                BuiltIn().log_to_console(
                    "3. Not found screenshot of expected result in baseline. Save new baseline with expected results")
                MegaBaseLine().save(screenshot_actual, remote_mega_dir_name)
                Artifacts()._delete_temp_dir(temp_dir)
        else:
            raise Exception(
                "Baseline with expected results is not set in visual tests. Please check configuration.")

    def _fail_visual_test(self, message,
                          baseline_screenshot_name,
                          path_to_joined_actual_expected_diff_image):
        """
        Fails visual test with message and joined screenshot (actual result / expected result / diff) in report.
        :param message: Text to be placed under the screenshot in report
        :param baseline_screenshot_name:
        :param path_to_joined_actual_expected_diff_image:
        """
        current_baseline_dir = Variables().get_variable_value(
            "${MEGA_BASELINE_DIR_NAME}")
        diff_image_file_name = os.path.split(
            path_to_joined_actual_expected_diff_image)[1]
        raise Exception(f'*HTML*Verification: <b>{message}</b>'
                        f'<br>Base line:'
                        f'<br> - directory: <i>"{current_baseline_dir}"</i>'
                        f'<br> - file: "<i>{baseline_screenshot_name}"</i>'
                        f'<br>ACTUAL / EXPECTED / DIFF:<br>'
                        f'<a href="{DIR_FOR_SCREENSHOTS_IN_REPORT}/{diff_image_file_name}">'
                        f'<img src="{DIR_FOR_SCREENSHOTS_IN_REPORT}/{diff_image_file_name}"'
                        f' width="98%" border="0" alt=""></a>')



from setuptools import setup

version = '0.2.5'

setup(name='VisualAssertLibrary',
      version=version,
      description='Robot Framework Visual Assert Library',
      packages=['VisualAssertLibrary'],
      author='Dmytro Bondarchuk',
      author_email='d.bondarchuk.spv@gmail.com',
      zip_safe=False)
